/* Emulator for an imaginery RISC processor
 * Author: Alexandros Frantzis, 2003 
 * email: alf82 at freemail dot gr
 * Licence: GPL
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "vm.h"

#define SHORT(a,b) ((int)(a)|(((int)(b))<<8))


unsigned char rom[]="\x02\x00\x00\x00" // loadil r0,0
		    "\x03\x00\x00\x00" // r0=0
		    "\x01\x01\x00\x00" // r1=0
		    "\x02\x01\x01\x00" // r1=1
		    "\x0c\xdf\x03\x02" // PRINT "Welcome to Skynet:"

		    "\xff\x90\x00\x00" // // RANDOM 0-9
		    "\x04\x00\x01\x90"
		    "\x0c\x00\x01\x00" // PRINT Random Code
		    
		    "\xff\x91\x00\x00"
		    "\x04\x00\x01\x91"
		    "\x0c\x00\x01\x00" // PRINT Random Code
		    
		    "\xff\x92\x00\x00"
		    "\x04\x00\x01\x92"
		    "\x0c\x00\x01\x00" // PRINT Random Code

		    "\x02\x0a\x0a\x00" // Place 0x0a ('\n') at 0x000a
		    "\x04\x00\x01\x0a"
		    "\x0c\x00\x01\x01" // PRINT '\n'
		    
		    "\x0b\x00\x00\x00" // JUMP TO CODE 0x00
/* 0x03ec: */	    "Welcome to Skynet. Please sort: \0"
		    ;
/* encryption routine */
unsigned char enc[]="\xB9\x74\x09\x00\x00\xB8\xAC\x83\x04\x08\x80\x30\x55\x40\x49\x75\xF9\xE9\x00\x80\x04\x08";

/*
unsigned char alfcode[]="\x0d\x09\x91\x90" // setlt r9,r91,r90
		        "\x08\x09\x00\x01" // be r0,r0,+1
			"\x82\x91\x90\x00" // swap r91,r90
			"\x0d\x09\x92\x90" // setlt r9,r92,r90
		        "\x08\x09\x00\x01" // be r9,r0,+1
			"\x82\x90\x92\x00" // swap r92,r90
			"\x0d\x09\x92\x91" // setlt r9,r92,r91
		        "\x08\x09\x00\x01" // be r9,r0,+1
			"\x82\x91\x92\x00" // swap r92,r91
			
			
			"\x04\x00\x01\x90" 
		        "\x0c\x00\x01\x00" // PRINT 
			"\x04\x00\x01\x91"
			"\x0c\x00\x01\x00" // PRINT 
			"\x04\x00\x01\x92"
			"\x0c\x00\x01\x00" // PRINT 
			"\x00\x00\x00\x00" // halt
			;
*/
int main(int argc, char **argv)
{
    int done=0;
    vm *v;
    unsigned char b0,b1,b2,b3,t;
    int s;
    int rom_start,rom_size;
    FILE *fp=NULL;
    
    
    v=malloc(sizeof(*v));
    memset(v->mem,0,sizeof(v->mem));
    memset(v->reg,0,sizeof(v->reg));        
    rom_size=sizeof(rom)-1;
    rom_start=MEM_SIZE-rom_size;
    Debug("-- Debug Mode ON. Compile with '-DNDEBUG' to turn off debug messages\n");
    Debug("Loading ROM Image at 0x%x(%d) size: 0x%x(%d)\n",rom_start,rom_start,rom_size,rom_size);
    memcpy(v->mem+rom_start,rom,rom_size);
    v->ip=rom_start;
    
    /* read module */
    if (argc>1)
	fp=fopen(argv[1],"r");
    if (fp!=NULL) {
	int nread;
	nread=fread(v->mem,1,rom_start,fp);
	Debug("Opened sort module and read %d bytes into main memory\n",nread);
    }
    
    //memcpy(v->mem,alfcode,sizeof(alfcode)-1);
    //v->mem[0x00]='\x00'; //HALT

    srandom(time(NULL));
    
    while(!done) {
    
	/* if out of bounds restart */
	if ( (v->ip < 0) || (v->ip >= MEM_SIZE)) {
	    Debug("Resetting: IP out of bounds %x(%d)\n",v->ip,v->ip);
	    memcpy(v->mem+rom_start,rom,rom_size);
	    v->ip=rom_start;
	}

	/* fetch instruction bytes */
	b0=v->mem[v->ip++];
	b1=v->mem[v->ip++];
	b2=v->mem[v->ip++];
	b3=v->mem[v->ip++];
	
	/* switch on opcode */
	switch(b0) {
	    case HALT: /* 0x00 */
		    Debug("0x%x: Halt :(\n",v->ip-4);
		    done=1;
		    break;
	    case ADD: /* 0x01 */
		    v->reg[b1]=v->reg[b2]+v->reg[b3];
		    Debug("0x%x: R[%d]=R[%d]+R[%d]\n",v->ip-4,b1,b2,b3);
		    break;
		    
	    /* STORE, LOAD etc */    
	    case LOADIL: /* 0x02 */
		    v->reg[b1]= (v->reg[b1]&0xffff0000) | (SHORT(b2,b3) &0x0000ffff) ;
		    Debug("0x%x: Rlow[%d]=%d\n",v->ip-4,b1,SHORT(b2,b3));
		    break;
	    case LOADIH: /* 0x03 */
		    v->reg[b1]= (v->reg[b1]&0x0000ffff) | ((SHORT(b2,b3) & 0x0000ffff)<<16) ;
		    Debug("0x%x: Rhigh[%d]=%d\n",v->ip-4,b1,SHORT(b2,b3));
		    break;
	    case STORE: /* 0x04 */ 
		    v->mem[SHORT(b1,b2)]=v->reg[b3];
		    Debug("0x%x: M[%d]=R[%d]\n",v->ip-4,SHORT(b1,b2),b3);
		    break;
	    case LOAD: /* 0x05 */
		    v->reg[b3]=v->mem[SHORT(b1,b2)];
		    Debug("0x%x: R[%d]=M[%d]\n",v->ip-4,b3,SHORT(b1,b2));
		    break;
	    case INDSTORE: /* 0x06 */ 
		    v->mem[v->reg[b1]+v->reg[b2]]=v->reg[b3];
		    Debug("0x%x: *M[0x%x+0x%x]=R[%d]\n",v->ip-4,v->reg[b1],v->reg[b2],b3);
		    break;
	    case INDLOAD: /* 0x07 */
		    v->reg[b3]=v->mem[v->reg[b1]+v->reg[b2]];
		    Debug("0x%x: R[%d]= *M[0x%x+0x%x]\n",v->ip-4,b3,v->reg[b1],v->reg[b2]);
		    break;
		    
	    /* Branching */		    
	    case BE: /* 0x08 */
		    Debug("0x%x: if (R[%d]==R[%d]) branch to 0x0x%x...",v->ip-4,b1,b2,v->ip+4+(char)b3*4);
		    if (v->reg[b1]==v->reg[b2]) {
			Debug("Taken.");
			v->ip += (char)b3*4;
		    }
		    Debug("\n");			
		    break;
	    case BNE: /* 0x09 */
		    Debug("0x%x: if (R[%d]!=R[%d]) branch to 0x0x%x...",v->ip-4,b1,b2,v->ip+4+(char)b3*4);
		    if (v->reg[b1]!=v->reg[b2]) {
			Debug("Taken.");
			v->ip += (char)b3*4;
		    }
		    Debug("\n");			
		    break;
	    case BU: /* 0x0a */
		    Debug("0x%x: branch to 0x%x\n",v->ip-4,SHORT(b1,b2));	    
		    v->ip += SHORT(b1,b2)*4;
		    break;
	    case BR: /* 0x0b */
		    Debug("0x%x: branch register to 0x%x\n",v->ip-4,v->reg[b1]);	    
		    v->ip = v->reg[b1];
		    break;		    
	    case PRINT: /* 0x0c */
		    s=SHORT(b1,b2);
		    if (s<0 || s>=MEM_SIZE) {
			    memcpy(v->mem+rom_start,rom,rom_size);
			    v->ip=rom_start;
		    }	    	
		    else if (b3==0)
			printf("%x",v->mem[s]);
		    else if (b3==1)
			printf("%c",v->mem[s]);
		    else if (b3==2)
			printf("%s",&v->mem[s]);
		    break;
	    case SETLT: /* 0x0d */ 
		    if (v->reg[b2]<v->reg[b3])
			v->reg[b1]=1;
		    else
			v->reg[b1]=0;    
		    break;


	    case SWAP: /* 0x82 */	    
		    Debug("0x%x: swap R[%d]<=>R[%d]\n",v->ip-4,b1,b2);
		    t=v->reg[b1];
		    v->reg[b1]=v->reg[b2];
		    v->reg[b2]=t;
		    break;
	    case RANDOM: /* 0xff */ 
			Debug("0x%x: Random at R[%d]\n",v->ip-4,b1);
			v->reg[b1]=random()%10;
		    break;
    	    default:
		break;
	}
		
    }
    
}

